//
//  AerosolPhase.cpp
//  soot_stochastic
//
//  Created by Alexandre Bouaniche on 29/08/2017.
//  Copyright © 2017 Alexandre Bouaniche. All rights reserved.
//

#include "AerosolPhase.hpp"
#include "GasParticle.hpp"

#include <iostream>
#include <string>
#include <vector>
#include <cmath>

using namespace std;

AerosolPhase::AerosolPhase()
{
    vector<GasParticle> aerosolPhase;
    
    int npTot = 10000;
    
    double m(1.0);
    double c0(0.5);
    
    int i(0);
    for(i=0; i<npTot;i++)
    {
        m = m*1.001;
        GasParticle gasPart(m,c0);
        aerosolPhase.push_back(gasPart);
    }
    
    m_partVector = aerosolPhase;
}

AerosolPhase::AerosolPhase(Grid* gridPtr,int npTot)
{
    vector<GasParticle> allParticles;
    GridCell gridCell0 = gridPtr->getGridCell(0);
    
    double mInf0;
    double mSup0;
    
    mInf0  = gridCell0.getmInf();
    mSup0 = gridCell0.getmSup();
    
    int j(0);
    for(j=0; j<npTot; j++)
    {
        double randomMass = frand_a_b(mInf0, mSup0);
        allParticles.push_back(randomMass);
    }
    
}



AerosolPhase::AerosolPhase(Grid* gridPtr, double n0, double x0, int npTot, string initType)
{
    vector<GasParticle> allParticles;
    
    if(initType == "diameter")
    {
        cout << "WARNING AerosolPhase::AerosolPhase  initType = diameter not checked" << endl;
        int i(0);
        double dInfi;
        double dSupi;
        double dAvgi;
        
        int nBins = gridPtr->getSize();
        
        double initNvT(0.0);
        
        for(i=0; i<nBins; i++)
        {
            GridCell gridCelli = gridPtr->getGridCell(i);
            
            dInfi  = gridCelli.getdInfNano(); // diameter in nanometers
            dSupi = gridCelli.getdSupNano();
            dAvgi = gridCelli.getdAvgNano();
            
            double ndi = n0/x0 * exp(-dAvgi / x0);
            double dlog = log10(dSupi) - log10(dInfi);
            double nvi = ndi * dlog; // simple approximation (uniform distribution inside the sections)
            initNvT += nvi;
            
            
        }
        m_initNvT = initNvT;
        
        for(i=0; i<nBins; i++)
        {
            GridCell gridCelli = gridPtr->getGridCell(i);
            
            dInfi  = gridCelli.getdInfNano(); // diameter in nanometers
            dSupi = gridCelli.getdSupNano();
            dAvgi = gridCelli.getdAvgNano();
            
            double ndi = n0/x0 * exp(-dAvgi / x0);
            double dlog = log10(dSupi) - log10(dInfi);
            double nvi = ndi * dlog;   // simple approximation (uniform distribution inside the sections)
            
            int npi = rounding(nvi /initNvT *npTot);
            
            double randomMass(1.0);
            
            double mInfi = gridCelli.getmInf();
            double mSupi = gridCelli.getmSup();
            
            // for piecewise linear
            double nInfi = n0/x0 * exp(-dInfi / x0);
            double nSupi = n0/x0 * exp(-dSupi / x0);
            
            int j(0);
            for(j=0; j<npi; j++)
            {
                randomMass = piecewise_linear_distrib_n(mInfi, mSupi, nInfi, nSupi);
                //randomMass = frand_a_b(mInfi, mSupi);
                allParticles.push_back(randomMass);
            }
        }
    }
    else if(initType == "mass")
    {
        int i(0);
        double mInfi;
        double mSupi;
        double mAvgi;
        
        int nBins = gridPtr->getSize();
        
        double initNvT(0.0);
        
        // for exact mass conservation
        vector<double> theoreticalMass;
        vector<double> massCorrection;
        
        for(i=0; i<nBins; i++)
        {
            GridCell gridCelli = gridPtr->getGridCell(i);
            
            mInfi  = gridCelli.getmInf();
            mSupi = gridCelli.getmSup();
            mAvgi = gridCelli.getmAvg();
            
            //double ndmi = n0/x0 * exp(-mAvgi / x0);
            //double dm = mSupi - mInfi;
            //double nvi = ndmi * dm;    // simple approximation (uniform distribution inside the sections)
            // more accurate to calculate the analytical integral based on the analytical expression
            // n(x) = A * exp(B * x + C) with A = n0/x0   B = - 1 / x0    C = 0
            // primitive of n(x) is therefore   A / B * exp (Bx)
            double A = n0/x0;
            double B = - 1.0 / x0;
            double primitive_n_mSup = A / B * exp(B * mSupi);
            double primitive_n_mInf = A / B * exp(B * mInfi);
            double nvi = primitive_n_mSup - primitive_n_mInf;
            
            initNvT += nvi;
            
            // integration per parts used to find analytical expression of mSection (the integrated mass over the section, integral(x*n) ). see analysis.R
            double primitive_xn_mSup = primitive_n_mSup * (mSupi - 1.0 / B);
            double primitive_xn_mInf = primitive_n_mInf * (mInfi - 1.0 / B);
            double mSection = primitive_xn_mSup - primitive_xn_mInf;
            theoreticalMass.push_back(mSection);
            //cout << "theoretical mass in section " << i << " = " << mSection << endl;
        }
        
        m_initNvT = initNvT;
        
        for(i=0;i<nBins; i++)
        {
            GridCell gridCelli = gridPtr->getGridCell(i);
            
            mInfi  = gridCelli.getmInf();
            mSupi = gridCelli.getmSup();
            mAvgi = gridCelli.getmAvg();
            
            //double ndmi = n0/x0 * exp(-mAvgi / x0);
            //double dm = mSupi - mInfi;
            //double nvi = ndmi * dm;    // simple approximation (uniform distribution inside the sections)
            // more accurate to calculate the analytical integral based on the analytical expression
            // n(x) = A * exp(B * x + C) with A = n0/x0   B = - 1 / x0    C = 0
            // primitive of n(x) is therefore   A / B * exp (Bx)
            double A = n0/x0;
            double B = - 1.0 / x0;
            double primitive_n_mSup = A / B * exp(B * mSupi);
            double primitive_n_mInf = A / B * exp(B * mInfi);
            double nvi = primitive_n_mSup - primitive_n_mInf;
            
            int npi = rounding(nvi /initNvT *npTot);
            
            double randomMass(1.0);
            
            // for piecewise linear
            double nInfi = n0/x0 * exp(-mInfi / x0);
            double nSupi = n0/x0 * exp(-mSupi / x0);
            
            int j(0);
            for(j=0; j<npi; j++)
            {
                randomMass = piecewise_linear_distrib_n(mInfi, mSupi, nInfi, nSupi);
                //randomMass = frand_a_b(mInfi, mSupi);
                allParticles.push_back(randomMass);
            }
            
            // then make mass correction as in reallocation step
        }
    }
    else
    {
        cout << "ERROR initType" << endl;
    }
    
    m_partVector = allParticles;
}


AerosolPhase::AerosolPhase(string pathProject, string pathTarget, string dataFilename)
{
    vector<double> dataVector = readDataVectors(pathProject, pathTarget, dataFilename);
    vector<GasParticle> aerosolPhase;
    
    double c0(0.5);
    double m(0.0);
    int i(0);
    for(i=0; i<dataVector.size();i++)
    {
        m = dataVector[i];
        GasParticle gasPart(m,c0);
        aerosolPhase.push_back(gasPart);
    }
    
    m_partVector = aerosolPhase;
}


AerosolPhase::AerosolPhase(int npTot)
{
    vector<GasParticle> aerosolPhase;
    
    double m(1.0);
    double c0(0.5);
    
    int i(0);
    for(i=0; i<npTot;i++)
    {
        m = m*1.001;
        GasParticle gasPart(m,c0);
        aerosolPhase.push_back(gasPart);
    }
    
    m_partVector = aerosolPhase;
}


AerosolPhase::AerosolPhase(int npTot, std::vector<double> massVector)
{
    vector<GasParticle> aerosolPhase;
    
    double c0(0.5);
    double m(0);
    
    int i(0);
    for(i=0; i<npTot;i++)
    {
        m = massVector[i];
        GasParticle gasPart(m,c0);
        aerosolPhase.push_back(gasPart);
    }
    
    m_partVector = aerosolPhase;
}


AerosolPhase::AerosolPhase(int npTot, std::vector<double> massVector, std::vector<double> cVector)
{
    vector<GasParticle> aerosolPhase;
    
    double c(0.0);
    double m(0.0);
    
    int i(0);
    for(i=0; i<npTot;i++)
    {
        m = massVector[i];
        c = cVector[i];
        GasParticle gasPart(m,c);
        aerosolPhase.push_back(gasPart);
    }
    
    m_partVector = aerosolPhase;
}

// binInit
AerosolPhase::AerosolPhase(Grid* gridPtr, int npTot, int bin, double initNvT)
{
    vector<GasParticle> allParticles;
    
    GridCell gridCelli = gridPtr->getGridCell(bin);
    
    double mInfi  = gridCelli.getmInf();
    double mSupi = gridCelli.getmSup();
    
    double randomMass(mInfi);
    
    int j(0);
    for(j=0; j<npTot; j++)
    {
        randomMass = frand_a_b(mInfi, mSupi);
        
        allParticles.push_back(randomMass);
    }
    
    m_partVector = allParticles;
    m_initNvT = initNvT;
}


int AerosolPhase::getNpTot()
{
    int npTot(0);
    int i(0);
    for(i=0; i<m_partVector.size();i++)
    {
        npTot++;
    }
    
    return npTot;
}


GasParticle* AerosolPhase::getGasPartPtr(int i)
{
    GasParticle* ptr = &m_partVector[i];
    return ptr;
}


double AerosolPhase::getCi(int i)
{
    GasParticle* ptr = &m_partVector[i];
    double c = ptr->getC();
    return c;
}


void AerosolPhase::setCi(int i, double c)
{
    GasParticle* ptr = &m_partVector[i];
    ptr->setC(c);
}

double AerosolPhase::getMass(int i)
{
    GasParticle* ptr = &m_partVector[i];
    double mass = ptr->getMass();
    return mass;
}

void AerosolPhase::setMass(int i, double mass)
{
    GasParticle* ptr = &m_partVector[i];
    ptr->setMass(mass);
}

void AerosolPhase::addMass(int i, double deltaMass)
{
    GasParticle* ptr = &m_partVector[i];
    ptr->addMass(deltaMass);
}

/*
  // not used anymore. count done directly in Psd beacause mAvgVar is now calculated at the same time
int AerosolPhase::countNpl(double mInf, double mSup)
{
    double npl(0);
    int i(0);
    for(i=0; i<m_partVector.size();i++)
    {
        double m = getMass(i);
        if(m >= mInf  &&  m < mSup)
        {
            npl++;
        }
    }
    return npl;
}
 */


double AerosolPhase::getInitNvT()
{
    return m_initNvT;
}


vector<GasParticle*> AerosolPhase::getRankedPartPtr()
{
    vector<GasParticle*> rankedPartPtrVec;
    int i(0); // i rank in m_partVector
    
    GasParticle* partPtr = getGasPartPtr(0);
    rankedPartPtrVec.push_back(partPtr);
    
    for(i=1;i<m_partVector.size();i++)
    {
        GasParticle* partPtr = getGasPartPtr(i);
        double mi = partPtr->getMass();
        
        int rank(0);
        
        int j(0);  // rank goes through rankedPartPtrVec to get the rank at which the new partPtr is going to be inserted
        
        double mjMinusOne = mi / 2.0; // arbitrary value just to initiate the while loop
        while(mi>mjMinusOne && j < rankedPartPtrVec.size())
        {
            mjMinusOne = rankedPartPtrVec[j]->getMass();
            j++;
        }
        
        if(mi>mjMinusOne)  // case out of the loop because reached the end of rankedPartPtrVec.size()
        {
            rank = j;
        }
        else
        {
            rank = j-1;
        }
        
        vector<GasParticle*>::iterator iter;
        iter = rankedPartPtrVec.begin();
        
        rankedPartPtrVec.insert(iter+rank,partPtr);
        
    }
    
    return rankedPartPtrVec;
}
 
