//
//  GrowthTensor.cpp
//  npsSolver
//
//  Created by Alexandre Bouaniche on 16/02/2018.
//  Copyright © 2018 Alexandre Bouaniche. All rights reserved.
//

#include "GrowthTensor.hpp"

using namespace std;

GrowthTensor::GrowthTensor(Psd* psdPtr, double G0, double teta, double Malpha, string solverType, string sectionalMethod): m_psdPtr(psdPtr), m_G0(G0), m_teta(teta), m_Malpha(Malpha), m_solverType(solverType), m_sectionalMethod(sectionalMethod), m_growthtStep(0.0)
{
    setgGasTensor();
    setgOutTensor();
    setgOutNegTensor();
}


void GrowthTensor::setgGasTensor()
{
    vector<double> gGasTensor;
    int i(0);
    int nBins = m_psdPtr->getSize();
    for(i=0;i<nBins;i++)
    {
        double mInf = m_psdPtr->getmInf(i);
        double mSup = m_psdPtr->getmSup(i);
        double gGasCoef = m_G0* ( pow(mSup, (m_teta+1.0)) / (m_teta+1.0)  -  pow(mInf, (m_teta+1.0)) / (m_teta+1.0) );
        gGasTensor.push_back(gGasCoef);
    }
    
    m_gGasTensor = gGasTensor;
}

void GrowthTensor::setgOutTensor()
{
    vector<double> gOutTensor;
    int i(0);
    int nBins = m_psdPtr->getSize();
    for(i=0;i<nBins;i++)
    {
        double mInf = m_psdPtr->getmInf(i);
        double mSup = m_psdPtr->getmSup(i);
        
        double a = 1 - m_Malpha / (mSup - mInf);
        double Mj_jPlusOne = mSup - m_Malpha / 2.0;
        
        double gOutCoef = (1.0 - a) * (m_Malpha + Mj_jPlusOne) / m_Malpha;
        gOutTensor.push_back(gOutCoef);
    }
    
    m_gOutTensor = gOutTensor;
}

double GrowthTensor::getDotmGasBin(int i)
{
    double mInf = m_psdPtr->getmInf(i);
    double mSup = m_psdPtr->getmSup(i);
    
    Bin* binPtr = m_psdPtr->getBinPtr(i);
    double NvResiduali = binPtr->getNvResidual();
    double nmResiduali = NvResiduali / (mSup - mInf);   // [part.cm-3.x-1]
    
    double dotmGasBin(0.0);
    
    if(m_sectionalMethod == "twoPoints")
    {
        dotmGasBin = nmResiduali * m_gGasTensor[i];  // gCoef in [g.part-1.s-1.x]  -> dotmGasBin in [g.cm-3.s-1]
    }
    else if(m_sectionalMethod == "2pointGdt")
    {
        dotmGasBin = nmResiduali * gGasCoef(i);   // here gGasCoef depends on G*dt. Calculated at each time step different from above m_gGasTensor which depends only on fixed Malpha
    }
    else
    {
        cout << "ERROR sectional method in GrowthTensor::getDotmGasBin" << endl;
    }
    
    
    return dotmGasBin;
}

double GrowthTensor::getDotmOutBin(int i)
{
    double dotmOutBin(0.0);
    
    if(m_sectionalMethod == "twoPoints")
    {
        dotmOutBin = getDotmGasBin(i) * m_gOutTensor[i];  // gOutCoef in [] -> dotmOutBin in [g.cm-3.s-1]
        
        // dotmOutBin positive if G0 positive
    }
    else if(m_sectionalMethod == "2pointGdt")
    {
        dotmOutBin = getDotmGasBin(i) * gOutCoef(i);  // here gOutCoef depends on G*dt. Calculated at each time step different from above m_gOutTensor which depends only on fixed Malpha
    }
    else
    {
        cout << "ERROR sectional method in GrowthTensor::getDotmGasBin" << endl;
    }
    
    
    return dotmOutBin;
}

double GrowthTensor::getDotmTotBin(int i)
{
    int nBins = m_psdPtr->getSize();
    double dotmTotBin(0.0);
    
    if(m_G0 >= 0)
    {
        if(i==0)
        {
            dotmTotBin = getDotmGasBin(i) - getDotmOutBin(i);  // convention: substract positive value (- valNeg)
        }
        else if(i<(nBins-1))
        {
            dotmTotBin = getDotmGasBin(i) - getDotmOutBin(i) + getDotmOutBin(i-1);
        }
        else if(i==(nBins-1))
        {
            dotmTotBin = getDotmGasBin(i) + getDotmOutBin(i-1);
        }
        else
        {
            dotmTotBin = 0.0;
            cout << "ERROR dotmTotBin" << endl;
        }
        
    }
    else
    {
        if(i==0)
        {
            dotmTotBin = getDotmGasBin(i) - getDotmOutNegBin(i+1);  // convention: in the case G0 <0 the terms of the equation are negative so that the mass transfer from "right" (from j+1) is positive as it is substracted
        }
        else if(i<(nBins-1))
        {
            dotmTotBin = getDotmGasBin(i) + getDotmOutNegBin(i) - getDotmOutNegBin(i+1);
        }
        else if(i==(nBins-1))
        {
            dotmTotBin = getDotmGasBin(i) + getDotmOutNegBin(i);
        }
        else
        {
            dotmTotBin = 0.0;
            cout << "ERROR dotmTotBin" << endl;
        }
    }
    
    return dotmTotBin;
}

double GrowthTensor::getDotNvBin(int i)
{
    double mAvg = m_psdPtr->getmAvg(i);
    double dotNvBin(0.0);
    double nBins = m_psdPtr->getSize();
    
    //double mInf = m_psdPtr->getmInf(i);
    //double mSup = m_psdPtr->getmSup(i);
    
    
    if(m_sectionalMethod == "twoPoints")
    {
        dotNvBin = getDotmTotBin(i) / mAvg; // [g.cm-3.s-1] / [g/part] -> [part.cm-3.s-1]
    }
    else if(m_sectionalMethod == "2pointGdt")
    {
        //double Madd = getG(i) * m_growthtStep;  // difference with fixed Malpha. Madd depends on time step and G
        //double a = 1 - Madd / (mSup - mInf);
        //double Mj_jPlusOne = mSup - Madd / 2.0;  // faire une fonction qui calcule Mj_jPlusOne en fonction de i et l'injecter dans dotNv en dessous en remplaçant mAvg là où il faut.  !!!!!!!!!!!!!!!!!!!
        
        
        if(m_G0 >= 0)
        {
            if(i==0)
            {
                dotNvBin = getDotmGasBin(i)/mAvg - getDotmOutBin(i) / mAvg;  // convention: substract positive value (- valNeg)
                
                //dotNvBin = omega_bini(i) / mAvg; // negative term included
                
                //dotNvBin = a* Madd *omega_r / mAvg  - (1-a) * Mj_jPlusOne * omega_r / Mj_jPlusOne;
                
                //dotNvBin =  - (1-a) * Mj_jPlusOne * omega_r(i) / Mj_jPlusOne;
            }
            else if(i<(nBins-1))
            {
                
                dotNvBin =  getDotmGasBin(i)/mAvg - getDotmOutBin(i)/ mAvg + getDotmOutBin(i-1)/ mAvg;
                
                
                //dotNvBin = omega_bini(i) / mAvg + omega_biniPlusOne(i-1) /mAvg;
                
                //dotNvBin = a* Madd *omega_r / mAvg - (1-a) * Mj_jPlusOne * omega_r / Mj_jPlusOne + omega_biniPlusOne(i-1) / (Madd + mj_jPlusOne(i-1)) ;
                
                //dotNvBin = - (1-a) * Mj_jPlusOne * omega_r(i) / Mj_jPlusOne + omega_biniPlusOne(i-1) / (Madd + mj_jPlusOne(i-1)) ;
            }
            else if(i==(nBins-1))
            {
                
                dotNvBin = getDotmGasBin(i)/mAvg + getDotmOutBin(i-1)/ mAvg;
                
                //dotNvBin = omega_bini(i) / mAvg + omega_biniPlusOne(i-1)/mAvg;
                
                //dotNvBin = a* Madd *omega_r / mAvg  - (1-a) * Mj_jPlusOne * omega_r / Mj_jPlusOne + omega_biniPlusOne(i-1) / (Madd + mj_jPlusOne(i-1)) ;
                
                //dotNvBin = - (1-a) * Mj_jPlusOne * omega_r(i) / Mj_jPlusOne + omega_biniPlusOne(i-1) / (Madd + mj_jPlusOne(i-1)) ;
            }
            else
            {
                dotNvBin = 0.0;
                cout << "ERROR GrowthTensor::getDotNvBin" << endl;
            }
            
        }
        else
        {
            if(i==0)
            {
                //dotNvBin = getDotmGasBin(i)/mAvg - getDotmOutNegBin(i+1) / mAvg;  // convention: in the case G0 <0 the terms of the equation are negative so that the mass transfer from "right" (from j+1) is positive as it is substracted
                
                // to write later if > 0 works.
            }
            else if(i<(nBins-1))
            {
                //dotNvBin =  getDotmGasBin(i)/mAvg + getDotmOutNegBin(i)/ mAvg - getDotmOutNegBin(i+1)/ mAvg;
                // to write later if > 0 works.
            }
            else if(i==(nBins-1))
            {
                //dotNvBin = getDotmGasBin(i)/mAvg + getDotmOutNegBin(i)/ mAvg;
                // to write later if > 0 works.
            }
            else
            {
                dotNvBin = 0.0;
                cout << "ERROR GrowthTensor::getDotNvBin" << endl;
            }
        }
    }
    else
    {
        cout << "ERROR sectional method in GrowthTensor::getDotNvBin" << endl;
    }
    
    return dotNvBin;
}

double GrowthTensor::getDotmOutNegBin(int i)
{
    double dotmOutBin = getDotmGasBin(i) * m_gOutNegTensor[i];  // gOutCoef in [] -> dotmOutBin in [g.cm-3.s-1]
    
    // dotmOutBin negative if G0 positive
    
    return dotmOutBin;
}



void GrowthTensor::setgOutNegTensor()
{
    vector<double> gOutTensor;
    int i(0);
    int nBins = m_psdPtr->getSize();
    for(i=0;i<nBins;i++)
    {
        double mInf = m_psdPtr->getmInf(i);
        double mSup = m_psdPtr->getmSup(i);
        double Mox = m_Malpha;
        
        double a = 1 - Mox / (mSup - mInf);  // here as we use analytical expressions, we can use Mox = Malpha
        double Mj_jMinusOne = Mox / 2.0 + mInf;
        
        double gOutCoef = (1.0 - a) * (Mj_jMinusOne - Mox) / Mox;
        gOutTensor.push_back(gOutCoef);
    }
    
    m_gOutNegTensor = gOutTensor;
}

double GrowthTensor::getG(int i)
{
    double G(0.0);
    if(m_solverType == "analytical")
    {
        double mAvg = m_psdPtr->getmAvg(i);
        G = m_G0 * pow(mAvg, m_teta);
    }
    else
    {
        cout << "code not written in GrowthTensor::getG(int i) for solverType not analytical" << endl;
    }
    return G;
}

void GrowthTensor::setGrowthtStep(double growthtStep)
{
    m_growthtStep = growthtStep;
}

double GrowthTensor::gGasCoef(int i)
{
    double mSup = m_psdPtr->getmSup(i);
    double mInf = m_psdPtr->getmInf(i);
    double gasCoef = m_G0* ( pow(mSup, (m_teta+1.0)) / (m_teta+1.0)  -  pow(mInf, (m_teta+1.0)) / (m_teta+1.0) );
    return gasCoef;
}

double GrowthTensor::gOutCoef(int i)
{
    double mInf = m_psdPtr->getmInf(i);
    double mSup = m_psdPtr->getmSup(i);
    double Madd = getG(i) * m_growthtStep;  // difference with fixed Malpha. Madd depends on time step and G
    
    double a = 1 - Madd / (mSup - mInf);
    double Mj_jPlusOne = mSup - Madd / 2.0;
    
    double outCoef = (1.0 - a) * (Madd + Mj_jPlusOne) / Madd;
    return outCoef;
}

double GrowthTensor::gOutNegCoef(int i)
{
    double mInf = m_psdPtr->getmInf(i);
    double mSup = m_psdPtr->getmSup(i);
    double Mox =  abs(getG(i) * m_growthtStep);  // difference with fixed Malpha. Mox is positive by convention even if G is negative
    
    double a = 1 - Mox / (mSup - mInf);  // here as we use analytical expressions, we can use Mox = Malpha
    double Mj_jMinusOne = Mox / 2.0 + mInf;
    
    double outNegCoef = (1.0 - a) * (Mj_jMinusOne - Mox) / Mox;
    
    return outNegCoef;
}


double GrowthTensor::omega_bini(int i)
{
    double mSup = m_psdPtr->getmSup(i);
    double mInf = m_psdPtr->getmInf(i);
    
    double Madd = getG(i) * m_growthtStep;  // difference with fixed Malpha. Madd depends on time step and G
    double a = 1 - Madd / (mSup - mInf);
    double Mj_jPlusOne = mSup - Madd / 2.0;  // faire une fonction qui calcule Mj_jPlusOne en fonction de i et l'injecter dans dotNv en dessous en remplaçant mAvg là où il faut.  !!!!!!!!!!!!!!!!!!!
    double nm = m_psdPtr->calculateNm(i);
    
    double integralGnxdx = nm * gGasCoef(i);
    double omega_r = integralGnxdx / Madd;
    
    // DLR supplemental material
    double omega_bini = (a*Madd - (1-a) * Mj_jPlusOne) * omega_r;
    
    return omega_bini;
}

double GrowthTensor::omega_biniPlusOne(int i)
{
    double mSup = m_psdPtr->getmSup(i);
    double mInf = m_psdPtr->getmInf(i);
    
    double Madd = getG(i) * m_growthtStep;  // difference with fixed Malpha. Madd depends on time step and G
    double a = 1 - Madd / (mSup - mInf);
    double Mj_jPlusOne = mSup - Madd / 2.0;  // faire une fonction qui calcule Mj_jPlusOne en fonction de i et l'injecter dans dotNv en dessous en remplaçant mAvg là où il faut.  !!!!!!!!!!!!!!!!!!!
    double nm = m_psdPtr->calculateNm(i);
    
    double integralGnxdx = nm * gGasCoef(i);
    double omega_r = integralGnxdx / Madd;
    
    // DLR supplemental material
    double omega_biniPlusOne = (1-a)* (Madd +Mj_jPlusOne) * omega_r;
    return omega_biniPlusOne;
}

double GrowthTensor::mj_jPlusOne(int i)
{
    double mSup = m_psdPtr->getmSup(i);
    
    double Madd = getG(i) * m_growthtStep;  // difference with fixed Malpha. Madd depends on time step and G

    
    double Mj_jPlusOne = mSup - Madd / 2.0;
    return Mj_jPlusOne;
}

double GrowthTensor::omega_r(int i)
{
    double Madd = getG(i) * m_growthtStep;  // difference with fixed Malpha. Madd depends on time step and G
    
    double nm = m_psdPtr->calculateNm(i);
    
    double integralGnxdx = nm * gGasCoef(i);
    double omega_r = integralGnxdx / Madd;
    return omega_r;
}
