//
//  SootSourceTerm.hpp
//  soot_stochastic
//
//  Created by Alexandre Bouaniche on 04/09/2017.
//  Copyright © 2017 Alexandre Bouaniche. All rights reserved.
//

#ifndef SootSourceTerm_hpp
#define SootSourceTerm_hpp

#include <stdio.h>

#include <vector>
#include <string>

#include "HomogeneousGasPhase.hpp"
#include "Psd.hpp"
#include "AgglomerationTensor.hpp"

class SootSourceTerm
{
public:
    SootSourceTerm(HomogeneousGasPhase* gasPhasePtr, Psd* psdPtr, std::string solverType);
    
    SootSourceTerm(Psd* psdPtr, std::string solverType); // without pointer at HomogeneousGasPhase. For analytical cases
    
    // here we did not put the timePerIt value in the source terms like in previous version (there was *timePerIt in beta for example). Instead, all source terms will be calculated in [s^-1] and will need to be multiplied by timePerIt in S2solver

    // UNITS: [ g, mol or particles (Navogadro linking), cm, s, K ]  !!
    
        // The chosen mass unit is g. the mass of 1 carbon atom Cmass = 12.0 AMU. 1 AMU = 1.66e-24 g. Then 1 Cmass = 1.99e-23 g  and 1 mol of Cmass = 12.00 g
    
        // The particles and PAH diameters are calculated (in cm) from the mass considering spheres with rhoSoot = 1.8 g/cm^3
    
        // gas phase species concentrations are given in [mol/cm^3], particles Nv in [part/cm^3]. concentration * Navogadro = Nv   :    [mol/cm^3]  *  [part/mol] = [part/cm^3]
    
        // Nucleation and condensation modeling from KIT Blanquart, Pitsch, ~p476
        // beta values initially from KIT Mauss, Netzell p465 but then epsilonCond also = 2.2 (cf Frenklach_Wang sympo 1991
    
        // gamma nucl = 1.0 cf Frenklach_Wang sympo 1991 Detailed modeling of soot particle nucleation and growth
    
        // HACA and oxidation as in APPEL, BOCKHORN, FRENKLACH, 2000: Kinetic Modeling of Soot Formation with Detailed Chemistry and Physics: Laminar Premixed Flames of C2 Hydrocarbons. Combust. Flame 121:122 (2000).
    
    
    
    //main public functions (information needed for advancing the solver)
    
    double partNuclSource(double t);        // dotH
    double partDimerSource (double t);      // dotDimer. PAH pyrene C16H10   // [part/s/cm^3]
    std::vector<double> partAggloSourceVector(double t);  // dotAlVector. + exponential nucleation in if m_nuclType == "exponential"
    
    std::vector<double> deltamTotGrowthVector(double t);
    double deltamTotGrowthPart(double t,double m); // for NpSolver::advanceGrowthStochasticPart(). apply surf growth on particles with source terms calculated for each particle. For now only for analytical cases. In order to extend to HomogeneousPhase cases, betaCond(part), alpha(part) and surf(part) will have to be calculated instead of betaCond(bin), alpha(bin) and surf(bin)
    
    
        // main sub routines
    double deltamCondi(double t, int i);    // in [g/part/s]
    // from partCondSourcei, deduced mass of dimers condensated on particles of bini. Then calculates average mass added per "real" particle in bini. this mass is deltamCondi and will be added to all stochastic particles in bin i.
    
    double deltamSurfi(double t, int i);   // in [g/part/s]
    // converts surfMassSource in a deltaM per real particle in bin i
    
    
    
    // Condensation subroutines
    double dimerConcentration(double t);   // [part/cm3]  Solves QSS. Solution of a quadratic equation.
    double partCondSourcei(double t, int nini);  // [part/cm3/s]. Gives the number of Dimers per second condensated on particles of a Bin i. Then in another function, "translated" into a mass deltamCondi applied to each particle of bin i.
    double partCondSourceTot(double t);  // [molecPAH/cm3/s]
    
    
    // Surface reactions subroutines
    double surfMassSource(double t, int bini);   // in [g/cm3/s] for bin i.
    
    double concentrationActiveSites(double t, int bini);  // [mol of active sites / cm3]  [Ai] + [Ri]
    // from PhD thesis D'AubagnacKarkar (global logic for getting concentration) p 50 and APPEL, BOCKHORN, FRENKLACH, 2000 (constant CHI = 2.3e15) ~ lambda)  alpha expression tanh... ~ alpha = 0.27 in D'AubagnacKarkar. avg soot surface in a bin calculated as in GridCell object.
    double concentrationRadicalSites(double t, int bini);  // [mol of radical sites / cm3] [Ri]
    
    double kCoefReactionRates(double A, double n, double E, double t);
    
    //Pyrene consumption correction
    void addConsoTotPAH(double timePerIt, double t);  // in [g/cm3]
    double YA4corrected(double t);  // []
    double YA4hybrid(double t);  // YA4corrected until t = 5.2e-3 then, YA4imposed
    
    double getDotCmA4correctedSource(double t); // for "PAHsources" A4correctionType. Source term calculated in SootSourceTerm in [g/cm3/s]
    
    void advanceCmA4correctedSource(double wmA4perIt); // in NpSolver, the corresponding deltam during timePerIt is calculated and with advanceYA4correctedSource, is added to m_YA4correctedSource
    
    
    
    //The source terms for nucleation and agglomeration are in particles above   (alphaVector calculated in npSolver because need to have nT(t) and nT(t-dt) )
    // the source terms for surface growth are in mass as they are directly added on each particle.
    
    
    double getAlphaHacaUniform();
    double getCmA4();
    double alphaHacaBin(int bini, double t);
    double avgalphaHaca(double t);
    double kfInterpolation(int bini);
    double collisionDiameterCalc(double mass); // mass in [g], dc in [cm]
    
    // model configuration
    void setAggloType(std::string aggloType);
    void setNuclType(std::string nuclType);
    void setGrowthType(std::string growthType);
    
    void setPAHType(std::string PAHType);   // "nofeedback", "corrected", "hybrid"
    void setA4correctionType(std::string correctionType); // "consoTot", "PAHsources"
    
    void setHacaType(std::string hacaType);
    void setCondType(std::string condType);
    void setOxiO2Type(std::string oxiO2Type);
    void setOxiOHType(std::string oxiOHType);
    void setSurfType(std::string surfType);
    void setMassTransferType(std::string massTransferType);
    
    void setbetaConst(double betaConst);
    void setgammaNucl(double gammaNucl);
    void setg0(double g0);
    void setTheta(double theta);
    void setb0(double b0);
    void setx0n(double x0n);
    void setalphaHaca(double alpha);
    void setKf(double kf);
    void setImposedDpSource(double dp);
    void setDf(double Df);
    void setAggloTensor(AgglomerationTensor* aggloTensorPtr);
    
    
    void showBoltzmann();
    double getConsoTotPAH();
    double getDotmPAHnucl();
    double getDf();
    double surfFromFractal(int bin);
    
    HomogeneousGasPhase* getGasPhasePtr();
    
    // sub-functions of deltamtotGrowth
    /*
    double deltamHaca(double t, double particleMass);
    double deltamCond(double t, double particleMass);
    double deltamOxiO2(double t, double particleMass);
    double deltamOxiOH(double t, double particleMass);
     */
    
    
         // beta: sub-functions of massAggloSource and massNuclSource 1
    double betaFree(double m1, double m2, double T);   // hypothesis of spheres. equivalence between mass/volume/diameter
    double betaFreeDc(double mi, double mj, double dc1, double dc2, double T); // calculated with the collision diameter (aggregates). m_ aggloType "collisionDiameter"
    // diameters must be given in [cm] !!
    double betaContinuumDc(double dc1, double dc2, double T, double P);  // dc in [cm], T in [K], P in [Pa]
    double betaHarmonicDc(double mi, double mj, double dc1, double dc2, double T,double P);
    
    double betaNucl(double mDimer, double T);
    double betaCond(double mDimer, double m1, double T);
    double betaCondDc(double mDimer, double m1, double dc1, double T);  // diameters must be given in [cm] !!
    //double betaContinuum(double m1, double m2, double T, double P, double sigma);
    //double betaTransition(double m1, double m2, double T, double P, double sigma);
    double freePath(double T, double P);  // Mean free path for air in [cm] (from AK, myf 1997) P in Pa; T in K
    double knudsen(double T, double P, double dc); // Mean free path and dc in [cm], P in Pa; T in K
    double cunningham(double T, double P, double dc);
    double dynVisco(double T);
    
protected:
    HomogeneousGasPhase* m_gasPhasePtr;
    Psd* m_psdPtr;
    AgglomerationTensor* m_aggloTensorPtr;
    
    std::string m_solverType; // "HomogeneousGasPhase", "analytical"
    
    std::string m_aggloType; // "standard", "betaConst", "collisionDiameter", "harmonicAvgDc", "sumKernel"
    std::string m_nuclType;  // "kraft"(Pyrene nucl and cond), "standard" (dimers of pyrene nucl and cond), "gammaNucl" (dimers of pyrene nucl and cond), "exponential" (analytical test case), "dotHconst"
    
    std::string m_growthType;  // "standard", "G0mTheta" (not necessary to precise for analytical cases, in these cases, always G0mTheta is used)
    
    std::string m_hacaType;  // "standard", "constant", "appel"  (default [] 0.27 from PhD thesis D'AubagnacKarkar p 50. Appel: dependence law on T. About 0.7 for the studied Zhao flame. from "Kinetic Modeling of Soot Formation with Detailed Chemistry and Physics: Laminar Premixed Flames of C2 Hydrocarbons"
    std::string m_surfType; // "sphere", "fractal"
    
    // "appelAndFactor" same as Appel but multiplied by a factor
    
    // in Appel: alpha = tanh(a/log(mu1) + b)  with a = 12.65 - 0.00563 * T; b =  -1.38 + 0.00068 * T  and mu1 the first size moment of the soot particle distribution from ref [6] of paper Appel which is Bokhorn book.
    
    //p180 says: "the first size moment gives the average PAH size, i.e. the average number of carbon atoms per PAH"
    
    //p182. mu1 = M1 / M0 = sum(miNi) / sum(Ni)
    // mi is mass represented by the number of carbons. therefore mu1 = average number of carbons per particle of soot = avgMass of one particle / cmass
    
    std::string m_PAHType;  // "nofeedback", "corrected", "hybrid"
    std::string m_A4correctionType;  // "consoTot", "PAHsources"
    std::string m_massTransferType;  // "geo2SPAMM", "aggloTensor", "kumar96"
    
    // these are not used for now but could be used to add different models
    std::string m_condType;
    std::string m_oxi02Type;
    std::string m_oxiOHType;
    
    
    double m_betaConst;
    double m_gammaNucl;
    double m_alphaHACA;  // !!To implement instead of static constant !! [] 0.27 from PhD thesis D'AubagnacKarkar p 50. Appel: dependence law on T. About 0.7 for the studied Zhao flame
    double m_kf;
    double m_imposedDpSource;
    double m_Df;
    
    double m_consoTotPAH;  // in [g/cm3] sum over time of nuclMassSource * timePerIt. used for YA4' corrected mass fraction after pyrene consumption
    double m_dotmPAHNucl;  // in [g/cm3/s]. used to calculate consoTot and then YA4corrected
    double m_CmA4corrected; // CmA4corrected initiated with YA4(t=0) then CmA4corrected advanced at each iteration with wmA4 source term (deduced from cantera results) and dotmPAHNucl, dotmPAHcond
    
    //just for test cases
    double m_g0;
    double m_theta;
    double m_b0;
    double m_x0n;
    
    //test case: Nucleation happens at all sizes and does not depend on HomogeneousPhase
    std::vector<double> testCaseNuclSourceVector();
    
    
    //constant values used for source terms calculation
    
    static const double avogadro;
    static const double cmass;
    static const double boltzmann;
    static const double rhoSoot;
    static const double pi;
    static const double epsilon;      //Van der Waals enhancement factor for nucl and agglo
    static const double epsilonCond;  //Van der Waals enhancement factor for condensation
    static const double chi;
    static const double Rconst;    //  [kcal/mol]
    
    static const double fR1_A;     //  [cm3 mol-1  s-1]  forward reaction
    static const double fR1_n;     //
    static const double fR1_E;     //  [kcal/mol]
    
    static const double rR1_A;     // reverse reaction
    static const double rR1_n;
    static const double rR1_E;
    
    static const double fR2_A;
    static const double fR2_n;
    static const double fR2_E;
    
    static const double rR2_A;
    static const double rR2_n;
    static const double rR2_E;
    
    static const double fR3_A;
    static const double fR3_n;
    static const double fR3_E;
    
    static const double fR4_A;
    static const double fR4_n;
    static const double fR4_E;
    
    static const double fR5_A;
    static const double fR5_n;
    static const double fR5_E;
    
    static const double fR6_A;
    static const double fR6_n;
    static const double fR6_E;
    
    //static const double C1;           // constants C1 and C2 used for Sutherland's law for continuum regime
    //static const double C2;
    
    
    
    //internal use object subfunctions
    
            // sub-functions for growth
    int findBin(double m);
    
    
    // massTransferType == "kumar96"
    double kumarAggloSource(int iDestination); // in [part/cm3/s] ref Kumar and Ramkrishna 1996: eq (36), (39) and (41) or Park 2004 eq (2), (3)
    double kumarAggloSource(int iDestination, double T, double P); // not written yet
    double kumarEta(int k, int j, int iDestination); // k <= j
    
    double betaAggloHomogeneousGasPhase(int k, int j, double T, double P);
    double betaAggloAnalytical(int k, int j);
    
    // massTransferType == "aggloTensor"
        // !!!  in AgglomerationTensor (based on D'Aubagnac) colliding particles are i and j, destination bin is k. In "geo2SPAMM" from Eberle, colliding particles are k and j. destination bin j and j+1.
    double tensorWmTotk(int kDestination, double T, double P);  // in [g/s/cm3]. equivalent to wmTotj of "geo2SPAMM". Sum of positive and negative source terms.
    double tensorWmTotk(int kDestination); // for solverType == "analytical". No dependence of beta on gas phase
    
        //tensorWmTotk subfunctions
    double tensorWmPosk(int kDestination, double T, double P);  // in [g/s/cm3]. source term of mass received from all bins into k
    double tensorWmNegi(int iOrigin, double T, double P); // source term of mass consumed from the "i" bin colliding with all j bins.
    double tensorWmNegj(int jOrigin, double T, double P); // only tensorWmNegi OR tensorWmNegj is used
    
    // for solverType == "analytical". No dependence of beta on gas phase
    double tensorWmPosk(int kDestination);  // in [g/s/cm3]. source term of mass received from all bins into k
    double tensorWmNegi(int iOrigin); // source term of mass consumed from the "i" bin colliding with all j bins.
    double tensorWmNegj(int jOrigin); // only tensorWmNegi OR tensorWmNegj is used
    
        //end massTransferType == "aggloTensor"
    
    
    // massTransferType == "geo2SPAMM": sub-functions of massAggloSource and massNuclSource 2
    double w_kj(int k, int j, double T, double P);  // "reaction rate" collision BINk + BINj -> nu_j*BINj + nu_(j+1)*BINj+1. Only for agglo. the conditional test of m_aggloType is done here. P in [Pa]
    double w_kj(int k, int j); // for solverType == "analytical". No dependence of beta on gas phase (T,P)
    
    double avgMbinkTojPlusOne(int k, int j);  // used to calculate the nu coefficients for the transfer from one bin to the superior one.
    double avgMbinjTojPlusOne(int k, int j);
    double avgMbinkToj(int k, int j);
    
    double aCoef(int k, int j);
    double nuj(int k, int j);
    double nujPlusOne(int k, int j);
    
    double wmNegRj(int k, int j, double T, double P);
    double wmNegRk(int k, int j, double T, double P);
    double wmPosRj(int k, int j, double T, double P);
    double wmPosPlusOneRj(int k, int jPlusOne, double T, double P);
    double wmNegJ(int j, double T, double P);
    double wmPosJ(int j, double T, double P);
    double wmPosJPlusOne(int j, double T, double P);
    double wmTotj(int j, double T, double P);
    
    // end massTransferType == geo2SPAMM
    
    
    
    
    
};

#endif /* SootSourceTerm_hpp */
