//
//  TimeStep.cpp
//  npsSolver
//
//  Created by Alexandre Bouaniche on 22/01/2018.
//  Copyright © 2018 Alexandre Bouaniche. All rights reserved.
//

#include "TimeStep.hpp"

using namespace std;

TimeStep::TimeStep(double alpha, SootSourceTerm* sourcePtr, Psd* psdPtr, string tStepType, double maxGcoef, string growthtStepType, string sectionalMethod, GrowthPark* growthParkPtr): m_alpha(alpha), m_sourcePtr(sourcePtr), m_psdPtr(psdPtr), m_tStepType(tStepType), m_maxGcoef(maxGcoef), m_growthtStepType(growthtStepType), m_sectionalMethod(sectionalMethod), m_growthParkPtr(growthParkPtr)
{
    
}

double TimeStep::aggloTimeStep(double nvT, double dotH, double dotAt, vector<double> dotAiVec, double tsGrowth)
{
    // new criteria derived by inequalities on sum( abs(delta_ni) ) over i
    double sumAbsAi(0.0);
    double sumAi(0.0);
    int i(0);
    int nBins = m_psdPtr->getSize();
    for(i = 0; i < nBins; i++)
    {
        double dotAi = dotAiVec[i];
        sumAbsAi += abs(dotAi);
        
        //cout << "dotAi(" << i << ") = " << dotAi << endl;
    }
    
    double tstep1(0.0);
    // determinate if variation of total NvT dominates or if transfer between sections dominates
    if( abs(dotH + dotAt) > sumAbsAi)
    {
        //int np_0 = m_psdPtr->getNp(0);
        //int NpTot = m_psdPtr->getNpT();
        //double ratio_0 = double(np_0) / double(NpTot);
        double N0 = m_psdPtr->getNv(0);
        double ratio_0 = N0/nvT;
        
        
        if( abs(dotH + dotAt) > 100.0 * sumAbsAi &  sumAbsAi < (0.01 * nvT) & ratio_0 > 0.99 )
        //if( abs(dotH + dotAt) > 100.0 * sumAbsAi &  (abs(dotH + dotAt) + sumAbsAi) < (0.01 * nvT) )
        {
            
            tstep1 = (1.0 + m_alpha) * nvT / (abs(dotH + dotAt) + sumAbsAi);
            
            //tstep1 = m_alpha * nvT / (abs(dotH + dotAt) + sumAbsAi);
            
            /*
            cout << "regime abs(dotH + dotAt) > 100.0 * sumAbsAi" << endl;
            cout << "abs(dotH + dotAt) = " << abs(dotH + dotAt) << "   sumAbsAi = " << sumAbsAi << "   abs(dotH + dotAt) + sumAbsAi = " << abs(dotH + dotAt) + sumAbsAi << endl;
            cout << "dotH = " << dotH << "   dotAt = " << dotAt << endl;
             */
            
        }
        else
        {
            //tstep1 = m_alpha * nvT / abs(dotH + dotAt);
            tstep1 = m_alpha * nvT / (abs(dotH + dotAt) + sumAbsAi);
            
            /*
            cout << "regime abs(dotH + dotAt) <= 100.0 * sumAbsAi" << endl;
            cout << "abs(dotH + dotAt) = " << abs(dotH + dotAt) << "   sumAbsAi = " << sumAbsAi << "   abs(dotH + dotAt) + sumAbsAi = " << abs(dotH + dotAt) + sumAbsAi << endl;
            */
        }
        
    } else {
        //double K = 2.0 * m_alpha * 1.0 / ( abs(dotH + dotAt) + sumAbsAi );
        
        //double denom = 1.0/K - (dotH + dotAt);
        //tstep1 = nvT / denom;
        
        //tstep1 = m_alpha * nvT / sumAbsAi;
        tstep1 = m_alpha * nvT / (abs(dotH + dotAt) + sumAbsAi);
        
        //cout << "abs(dotH + dotAt) = " << abs(dotH + dotAt) << "   sumAbsAi = " << sumAbsAi << "   abs(dotH + dotAt) + sumAbsAi = " << abs(dotH + dotAt) + sumAbsAi << endl;
    }
    
    
    
    
    
    double tstep = tstep1;
    
    // Could be implemented: calculate tstep2 with the upper limit in function of tsGrowth (maximum allowed ratio)
    // then take the minimum of both tsteps
    
    // max_ratio * maxGcoef < 0.4 (CFL criteria)
    double max_ratio = 0.4 / m_maxGcoef;
    double tstep_ceil = max_ratio * tsGrowth;
    if(tstep > tstep_ceil)
    {
        tstep = tstep_ceil;
    }
    
    
    return tstep;
}

double TimeStep::aggloTimeStep(double t)
{
    double nvT = m_psdPtr->getNvT();
    double dotH = m_sourcePtr->partNuclSource(t);
    double dotAt = partAggloSourceTot(t);
    vector<double> dotAiVec = m_sourcePtr->partAggloSourceVector(t);
    
    double tsGrowth = growthTargetTimeStep(t);
    
    //double timeStep = aggloTimeStep(nvT, dotH, dotAt, tsPrevious);
    double timeStep = aggloTimeStep(nvT, dotH, dotAt, dotAiVec, tsGrowth);
    
    return timeStep;
}

double TimeStep::partAggloSourceTot(double t)
{
    double dotAt(0.0);
    vector<double> partAggloSource = m_sourcePtr->partAggloSourceVector(t);
    
    int i(0);
    for(i=0;i<partAggloSource.size();i++)
    {
        dotAt+=partAggloSource[i];
        //cout << "section: " << i << "   partAggloSource = " << partAggloSource[i] << endl;
    }
    
    return dotAt;
}

string TimeStep::tStepType()
{
    return m_tStepType;
}


double TimeStep::growthTargetTimeStep(double t)
{
    double tstep(0.0);
    
    if(m_growthtStepType == "growthStochastic")
    {
        tstep = growthTargetTsStochastic(t);
    }
    else if(m_growthtStepType == "growthSectional")
    {
        tstep = growthTargetTsSectional(t);
    }
    
    return tstep;
}



double TimeStep::growthTargetTsStochastic(double t)
{
    vector<double> deltamGrowthVector = m_sourcePtr->deltamTotGrowthVector(t);
    
    int nBins = m_psdPtr->getSize();
    
    // initialization with CFL criteria on last cell. Then test to take the minimum tstep on all cells CFL criteria.
    double mSupLast = m_psdPtr->getmSup(nBins-1);
    double mInfLast = m_psdPtr->getmInf(nBins-1);
    double deltamLast = mSupLast - mInfLast;
    
    double tStepMax = abs(m_maxGcoef * deltamLast / deltamGrowthVector[nBins-1]);
    
    double mintStep = tStepMax; // initialization
    
    int i(0);
    
    for(i=0;i<deltamGrowthVector.size();i++)
    {
        if(m_psdPtr->getNp(i)>0)
        {
            double mSupi = m_psdPtr->getmSup(i);
            double mInfi = m_psdPtr->getmInf(i);
            double deltami = mSupi - mInfi;
            
            double tStep = abs(m_maxGcoef * deltami / deltamGrowthVector[i]);
            if(tStep < mintStep)
            {
                mintStep = tStep;
            }
        }
        
    }
    
    return mintStep;
}



double TimeStep::growthTargetTsSectional(double t)
{
    int i(0);
    int nBins = m_psdPtr->getSize();
    
    // tStep initialized with CFL criteria at bin 0. Then compared to CFL criteria in all other bins and minimum is taken
    double mInf = m_psdPtr->getmInf(i);
    double mSup = m_psdPtr->getmSup(i);
    
    double Gestimated = m_sourcePtr->deltamTotGrowthVector(t)[i];
    
    
    // CFL criteria
    double deltam = mSup - mInf;
    double tStepMin = abs(m_maxGcoef * deltam / Gestimated);
    
    
    for(i=1;i<nBins;i++)
    {
        double mInf = m_psdPtr->getmInf(i);
        double mSup = m_psdPtr->getmSup(i);
        
        double Gestimated = m_sourcePtr->deltamTotGrowthVector(t)[i];
        
        
        // CFL criteria
        double deltam = mSup - mInf;
        double tStepBin = abs(m_maxGcoef * deltam / Gestimated);
        
        if(tStepBin < tStepMin)
        {
            tStepMin = tStepBin;
        }
    }
    
    return tStepMin;
    
}

int TimeStep::ratioTstepAggloGrowth(double t)
{
    //double tsAgglo = aggloTimeStep(t, tsPrevious);
    double tsAgglo = aggloTimeStep(t);
    double tsTargetGrowth = growthTargetTimeStep(t);
    double ratioDouble = tsAgglo / tsTargetGrowth;
    int ratio = ceil(ratioDouble);
    if(ratio < 1)
    {
        ratio = 1;
    }
    return ratio;
}

double TimeStep::finalGrowthTimeStep(double t)
{
    double tsAgglo = aggloTimeStep(t);
    int ratioInt = ratioTstepAggloGrowth(t);
    double ratioDouble = double(ratioInt);
    double tsGrowthFinal = tsAgglo / ratioDouble;
    return tsGrowthFinal;
}




